/*
 * Customizable Coaster - https://www.prusaprinters.org/prints/150339-customizable-coasters
 * by OutwardB          - https://www.prusaprinters.org/social/169322-outwardb/about
 * updated 2022-03-18
 * version v1.2
 *
 * ----------------------------
 *
 * This work is licensed under the Creative Commons - Attribution - Non-Commercial - ShareAlike license.
 * https://creativecommons.org/licenses/by-nc-sa/3.0/
 *
 * ----------------------------
 *
 * Changelog
 * --------------
 * V1.2
 *  - Added multi-color preview
 *
 * V1.1
 *  - Preview cup/coaster is no longer included in export
 *
 * V1
 *  - Initial release
 * 
 */

////////////////////////////////////////////////////
//////////////////////////////////////////////////// Parameter Section
////////////////////////////////////////////////////

/* [Export Settings] */
// Which part to export? (Use both for preview only!)
part_to_export = "coaster"; //[both,coaster,container]
// Show a preview cup/coasters? Useful for sizing.
enable_preview_items = "no"; //[yes,no]
// Size of the preview cup (diameter)
preview_cup_size = 72; //[10:100]

////////////////////////////////////////////////////

/* [Size and Shape Settings] */
// How to work out the size (Cup-Diameter is the inside diameter of cirle | Coaster-Diameter is the outside diameter of the circle | Coaster-Side-Length will work out the diameter from the side-length)
size_type = "cup-diameter"; //[cup-diameter,coaster-diameter,coaster-side-length]
// Set the size value
size_in_MM = 90; //[10:200]
// set to 21 for a circle
number_of_sides = 6; //[3:21]
// Higher number rounds the corners
corner_round = 3; //[0:30]

//////////////////////////////////////////////////// 

/* [Coaster Settings] */
// Height of the pattern (Use a multiple of your layer height for these values!)
coaster_pattern_height = 1.4; //[0.1:0.01:5]
// Base height (for Layer Change @ height)
coaster_base_height = 1.6; //[0:0.01:5]
// Raise outer border? (Border will be drawn on the inside of the coaster, make sure your cups will still fit (use the preview cup if needed)
extra_border_height = 2; //[0:0.01:5]
// For preview only! Use Coaster Pattern Type setting under Advanced Export Options to export the invert pattern
multicolor_preview = true;
// For preview only!
multicolor_preview_color = "Green";//[White,Black,Red,Blue,Green,Yellow,Orange,Pink,Purple,----Whites----,White,Snow,Honeydew,MintCream,Azure,AliceBlue,GhostWhite,WhiteSmoke,Seashell,Beige,OldLace,FloralWhite,Ivory,AntiqueWhite,Linen,LavenderBlush,MistyRose,----Grays----,Gainsboro,LightGrey,Silver,DarkGray,Gray,DimGray,LightSlateGray,SlateGray,DarkSlateGray,Black,----Reds----,IndianRed,LightCoral,Salmon,DarkSalmon,LightSalmon,Red,Crimson,FireBrick,DarkRed,----Blues----,Aqua,Cyan,LightCyan,PaleTurquoise,Aquamarine,Turquoise,MediumTurquoise,DarkTurquoise,CadetBlue,SteelBlue,LightSteelBlue,PowderBlue,LightBlue,SkyBlue,LightSkyBlue,DeepSkyBlue,DodgerBlue,CornflowerBlue,RoyalBlue,Blue,MediumBlue,DarkBlue,Navy,MidnightBlue,----Greens----,GreenYellow,Chartreuse,LawnGreen,Lime,LimeGreen,PaleGreen,LightGreen,MediumSpringGreen,SpringGreen,MediumSeaGreen,SeaGreen,ForestGreen,Green,DarkGreen,YellowGreen,OliveDrab,Olive,DarkOliveGreen,MediumAquamarine,DarkSeaGreen,LightSeaGreen,DarkCyan,Teal,----Yellows----,Gold,Yellow,LightYellow,LemonChiffon,LightGoldenrodYellow,PapayaWhip,Moccasin,PeachPuff,PaleGoldenrod,Khaki,DarkKhaki,Browns,Cornsilk,BlanchedAlmond,Bisque,NavajoWhite,Wheat,BurlyWood,Tan,RosyBrown,SandyBrown,Goldenrod,DarkGoldenrod,Peru,Chocolate,SaddleBrown,Sienna,Brown,Maroon,----Oranges----,LightSalmon,Coral,Tomato,OrangeRed,DarkOrange,Orange,----Pinks----,Pink,LightPink,HotPink,DeepPink,MediumVioletRed,PaleVioletRed,----Purples----,Lavender,Thistle,Plum,Violet,Orchid,Fuchsia,Magenta,MediumOrchid,MediumPurple,BlueViolet,DarkViolet,DarkOrchid,DarkMagenta,Purple,Indigo,DarkSlateBlue,SlateBlue,MediumSlateBlue]
// For preview only!
color_of_pattern = "Silver"; //[White,Black,Red,Blue,Green,Yellow,Orange,Pink,Purple,----Whites----,White,Snow,Honeydew,MintCream,Azure,AliceBlue,GhostWhite,WhiteSmoke,Seashell,Beige,OldLace,FloralWhite,Ivory,AntiqueWhite,Linen,LavenderBlush,MistyRose,----Grays----,Gainsboro,LightGrey,Silver,DarkGray,Gray,DimGray,LightSlateGray,SlateGray,DarkSlateGray,Black,----Reds----,IndianRed,LightCoral,Salmon,DarkSalmon,LightSalmon,Red,Crimson,FireBrick,DarkRed,----Blues----,Aqua,Cyan,LightCyan,PaleTurquoise,Aquamarine,Turquoise,MediumTurquoise,DarkTurquoise,CadetBlue,SteelBlue,LightSteelBlue,PowderBlue,LightBlue,SkyBlue,LightSkyBlue,DeepSkyBlue,DodgerBlue,CornflowerBlue,RoyalBlue,Blue,MediumBlue,DarkBlue,Navy,MidnightBlue,----Greens----,GreenYellow,Chartreuse,LawnGreen,Lime,LimeGreen,PaleGreen,LightGreen,MediumSpringGreen,SpringGreen,MediumSeaGreen,SeaGreen,ForestGreen,Green,DarkGreen,YellowGreen,OliveDrab,Olive,DarkOliveGreen,MediumAquamarine,DarkSeaGreen,LightSeaGreen,DarkCyan,Teal,----Yellows----,Gold,Yellow,LightYellow,LemonChiffon,LightGoldenrodYellow,PapayaWhip,Moccasin,PeachPuff,PaleGoldenrod,Khaki,DarkKhaki,Browns,Cornsilk,BlanchedAlmond,Bisque,NavajoWhite,Wheat,BurlyWood,Tan,RosyBrown,SandyBrown,Goldenrod,DarkGoldenrod,Peru,Chocolate,SaddleBrown,Sienna,Brown,Maroon,----Oranges----,LightSalmon,Coral,Tomato,OrangeRed,DarkOrange,Orange,----Pinks----,Pink,LightPink,HotPink,DeepPink,MediumVioletRed,PaleVioletRed,----Purples----,Lavender,Thistle,Plum,Violet,Orchid,Fuchsia,Magenta,MediumOrchid,MediumPurple,BlueViolet,DarkViolet,DarkOrchid,DarkMagenta,Purple,Indigo,DarkSlateBlue,SlateBlue,MediumSlateBlue]
// Only for preview - Base will only change color if it's actually enabled!
color_of_base = "Black"; //[White,Black,Red,Blue,Green,Yellow,Orange,Pink,Purple,----Whites----,White,Snow,Honeydew,MintCream,Azure,AliceBlue,GhostWhite,WhiteSmoke,Seashell,Beige,OldLace,FloralWhite,Ivory,AntiqueWhite,Linen,LavenderBlush,MistyRose,----Grays----,Gainsboro,LightGrey,Silver,DarkGray,Gray,DimGray,LightSlateGray,SlateGray,DarkSlateGray,Black,----Reds----,IndianRed,LightCoral,Salmon,DarkSalmon,LightSalmon,Red,Crimson,FireBrick,DarkRed,----Blues----,Aqua,Cyan,LightCyan,PaleTurquoise,Aquamarine,Turquoise,MediumTurquoise,DarkTurquoise,CadetBlue,SteelBlue,LightSteelBlue,PowderBlue,LightBlue,SkyBlue,LightSkyBlue,DeepSkyBlue,DodgerBlue,CornflowerBlue,RoyalBlue,Blue,MediumBlue,DarkBlue,Navy,MidnightBlue,----Greens----,GreenYellow,Chartreuse,LawnGreen,Lime,LimeGreen,PaleGreen,LightGreen,MediumSpringGreen,SpringGreen,MediumSeaGreen,SeaGreen,ForestGreen,Green,DarkGreen,YellowGreen,OliveDrab,Olive,DarkOliveGreen,MediumAquamarine,DarkSeaGreen,LightSeaGreen,DarkCyan,Teal,----Yellows----,Gold,Yellow,LightYellow,LemonChiffon,LightGoldenrodYellow,PapayaWhip,Moccasin,PeachPuff,PaleGoldenrod,Khaki,DarkKhaki,Browns,Cornsilk,BlanchedAlmond,Bisque,NavajoWhite,Wheat,BurlyWood,Tan,RosyBrown,SandyBrown,Goldenrod,DarkGoldenrod,Peru,Chocolate,SaddleBrown,Sienna,Brown,Maroon,----Oranges----,LightSalmon,Coral,Tomato,OrangeRed,DarkOrange,Orange,----Pinks----,Pink,LightPink,HotPink,DeepPink,MediumVioletRed,PaleVioletRed,----Purples----,Lavender,Thistle,Plum,Violet,Orchid,Fuchsia,Magenta,MediumOrchid,MediumPurple,BlueViolet,DarkViolet,DarkOrchid,DarkMagenta,Purple,Indigo,DarkSlateBlue,SlateBlue,MediumSlateBlue]

////////////////////////////////////////////////////

/* [Container Settings] */
// How many coasters should the container contain?
number_of_coasters = 4; //[2:20]
// How thick should the walls be?
container_wall_width = 2; //[1:0.01:20]
// Space between coasters and container walls
container_wall_buffer = 1; //[0.1:0.01:3]
// Height of the bottom base of the container
container_base_height = 1; //[0.1:0.01:3]
// How many slots
container_slot_type = "single"; //[single,double,alternating,all]
// How much space to keep in each corner between slots (inverted if using a circle)
container_slot_corner_buffer = 10; //[1:0.5:50]
// Curve of slot corner
container_slot_corner_val = 4; //[1:0.5:10]
// Only for preview
color_of_container = "Silver"; //[White,Black,Red,Blue,Green,Yellow,Orange,Pink,Purple,----Whites----,White,Snow,Honeydew,MintCream,Azure,AliceBlue,GhostWhite,WhiteSmoke,Seashell,Beige,OldLace,FloralWhite,Ivory,AntiqueWhite,Linen,LavenderBlush,MistyRose,----Grays----,Gainsboro,LightGrey,Silver,DarkGray,Gray,DimGray,LightSlateGray,SlateGray,DarkSlateGray,Black,----Reds----,IndianRed,LightCoral,Salmon,DarkSalmon,LightSalmon,Red,Crimson,FireBrick,DarkRed,----Blues----,Aqua,Cyan,LightCyan,PaleTurquoise,Aquamarine,Turquoise,MediumTurquoise,DarkTurquoise,CadetBlue,SteelBlue,LightSteelBlue,PowderBlue,LightBlue,SkyBlue,LightSkyBlue,DeepSkyBlue,DodgerBlue,CornflowerBlue,RoyalBlue,Blue,MediumBlue,DarkBlue,Navy,MidnightBlue,----Greens----,GreenYellow,Chartreuse,LawnGreen,Lime,LimeGreen,PaleGreen,LightGreen,MediumSpringGreen,SpringGreen,MediumSeaGreen,SeaGreen,ForestGreen,Green,DarkGreen,YellowGreen,OliveDrab,Olive,DarkOliveGreen,MediumAquamarine,DarkSeaGreen,LightSeaGreen,DarkCyan,Teal,----Yellows----,Gold,Yellow,LightYellow,LemonChiffon,LightGoldenrodYellow,PapayaWhip,Moccasin,PeachPuff,PaleGoldenrod,Khaki,DarkKhaki,Browns,Cornsilk,BlanchedAlmond,Bisque,NavajoWhite,Wheat,BurlyWood,Tan,RosyBrown,SandyBrown,Goldenrod,DarkGoldenrod,Peru,Chocolate,SaddleBrown,Sienna,Brown,Maroon,----Oranges----,LightSalmon,Coral,Tomato,OrangeRed,DarkOrange,Orange,----Pinks----,Pink,LightPink,HotPink,DeepPink,MediumVioletRed,PaleVioletRed,----Purples----,Lavender,Thistle,Plum,Violet,Orchid,Fuchsia,Magenta,MediumOrchid,MediumPurple,BlueViolet,DarkViolet,DarkOrchid,DarkMagenta,Purple,Indigo,DarkSlateBlue,SlateBlue,MediumSlateBlue]

////////////////////////////////////////////////////

/* [Pattern Settings] */
// Thickness of border in mm (Use a multiple of your line width setting in your slicer)
border_size = 2; //[0:0.02:55]
// Choose between a pattern
pattern_selection = "honeycomb"; //[aperture,bahtinov,bahtinov01,brick,brick-rounded,carey,cross,cross-ring,crosshair,crossshape,crossHearts,crosshex,crosstri,dot,fancy,fancy1,fancy2,grid,half-loop,honeycomb,line,masu-tsunagi,panel-squares,panel-squares-alt,random,rectangles,shippou-tsunagi,sin-waves,spiral,square,teardrop,triangle,trihole,tsutsuidutsu,web,windmill,zigzag,blank,none]
// Set the angle of the pattern around the center of the cover
pattern_rotation = 0; //[-180:0.01:180]
// Size of the pattern lines in mm. (Use a multiple of your line width setting in your slicer)
line_size = 2; //[0.3:0.02:20]
// Space between two lines of your pattern. If you choose aperture pattern, this value set the size of the center octagon.
line_space = 9; //[1:0.01:60]
// number of straight lines supporting some patterns (e.g. spiral, crosshair, square, etc )
number_of_support_lines = 6;  //[0:1:36]
// fill centre item for some patterns (crossheirs, etc)
centre_item = "yes"; //[yes,no,pattern]

////////////////////////////////////////////////////

/* [Text] */
// How are you customizing this thing? This will set the default font
What_app_are_you_using = "Thingiverse Customizer"; //[Thingiverse Customizer,OpenSCAD]
// Enable Text
text_enabled = "yes"; //[yes,no]
// Text to add
line_01 = "";
line_02 = "";
line_03 = "";
line_04 = "";
line_05 = "";

////////////////////////////////////////////////////

/* [Text Options] */
// Font for the grill text. Make sure all the text is attached! (per-line font in advanced options)
text_font = "";
// Font size of the grill text.
text_size = 8; //[0.5:0.5:100]
// Thickness of text outline in mm, set to 0 for a block item, 1 or more for an outline (Use a multiple of your line width setting in your slicer)
text_thickness = 1.2; //[0:0.1:10]
// "yes" to enable, "no" for auto-calculated value
enable_custom_line_space = "no"; //[yes,no]
// Extra space between text lines (enable above option to use this)
text_line_space = 1; //[-5:0.1:50]
// Cut the text from the pattern
cutout_text = "yes"; //[yes,no]
// Move text left/right
text_x_nudge = 0; //[-40:0.1:40]
// Move text up/down
text_y_nudge = 0; //[-40:0.1:40]
// Rotate text (Inverted if using a flat grill)
text_rotate = 0; //[-180:0.1:180]

////////////////////////////////////////////////////

/* [Advance Text Options] */
// Set to yes to enable a different font for each line
enable_multiple_fonts = "no"; //[yes,no]
// If blank, will use 'Text Font' value from -Text Options-
font_01 = "";
font_02 = "";
font_03 = "";
font_04 = "";
font_05 = "";

////////////////////////////////////////////////////

/* [Import SVG] */
// Set to yes to enable (Must be using OpenSCAD, will not work in Thingiverse customizer!)
svg_import_enabled = "no"; //[yes,no]
// Select the file to import (must be a SVG | Use forward slashes (even on windows) for example - C:/TempFolder/File.SVG )
file_to_import = "IIS.svg";
// Scale the SVG
file_scale = 1; //[0.001:0.001:50]
// Thickness of SVG outline, set to 0 for a block item, 1 or more for an outline
svg_thickness = 0; //[0:0.1:10]
// Move SVG left/right
svg_x_nudge = 0; //[-40:0.1:40]
// Move SVG up/down
svg_y_nudge = 0; //[-40:0.1:40]
// Rotate SVG
svg_rotate = 0; //[-180:0.1:180]
// Cut the SVG file from the pattern
cutout_svg = "yes"; //[yes,no]

////////////////////////////////////////////////////

/* [Advanced Export Options] */
// 3D or 2D render (2D works in OpenSCAD and for coaster only!)
2d_or_3d = "3D"; //[3D,2D]
// Basic for a normal pattern | Use inverted for multi-color/inlay | Set pattern to none to export text/svg only
coaster_pattern_type = "basic"; //[basic,inverted]
// Enable a custom skirt (makes it easier if inverted items are not centered)
multicolor_export_skirt = "no"; //[yes,no]
// Width of skirt (suggest using 3 * your line width setting in your slicer)
skirt_width = 1.2; //[0.1:0.01:3]
// Height of skirt (use 1 * your layer height)
skirt_height = 0.2; //[0.1:0.01:1]
// Distance between coaster and skirt. A few mm should be fine
skirt_buffer_size = 5; //[3:10]


////////////////////////////////////////////////////
//////////////////////////////////////////////////// Hidden Parameters
////////////////////////////////////////////////////
//[Hidden]

// preview[view:south]

rotate_val = isOdd(number_of_sides) ? 90 : 90 - (360/number_of_sides/2); // If number of sides is even, shape will be rotated so that the flat side is at the bottom

shape_radius =  size_type == "cup-diameter" ? radius_from_inradius(size_in_MM/2,number_of_sides):
                size_type == "coaster-side-length" ? radius_from_side_length(size_in_MM,number_of_sides):
                size_in_MM/2;

number_of_sides_val = number_of_sides == 21 ? size_in_MM : number_of_sides; // If more than 21 sides, set $fn of circle to double the size value

coaster_height = coaster_pattern_height + coaster_base_height + extra_border_height; // Calculate coaster height for container height

// Check how many lines are being used
used_line_count = text_enabled == "no" ? 0 :   
                    line_05 != "" ? 05 :
                    line_04 != "" ? 04 :
                    line_03 != "" ? 03 :
                    line_02 != "" ? 02 :
                    line_01 != "" ? 01 : 
                    0;

// Set text_set based on how many lines used
text_set =          
                    line_05 != "" ? [line_01, line_02, line_03, line_04, line_05] :
                    line_04 != "" ? [line_01, line_02, line_03, line_04] :
                    line_03 != "" ? [line_01, line_02, line_03] :
                    line_02 != "" ? [line_01, line_02] :
                    line_01 != "" ? [line_01] : 
                    [];
                    
pattern_text_font_val = text_font == "" ? What_app_are_you_using == "Thingiverse Customizer" ? "Allerta Stencil" : "Stencil" : text_font; // If no font is specified, use the default stenicl font based on app being used

// If using multiple fonts, check if each has a value and set to default if not
font_set = enable_multiple_fonts == "yes" ? [   font_01 == "" ? pattern_text_font_val : font_01, 
                                                font_02 == "" ? pattern_text_font_val : font_02, 
                                                font_03 == "" ? pattern_text_font_val : font_03, 
                                                font_04 == "" ? pattern_text_font_val : font_04, 
                                                font_05 == "" ? pattern_text_font_val : font_05] : 0;


coaster_d_value = 2d_or_3d == "3D" ? 3 : 2;

////////////////////////////////////////////////////
//////////////////////////////////////////////////// PROGRAM
////////////////////////////////////////////////////

// The main bit
if (part_to_export == "container") {
    color(color_of_container) container ();
} else if (part_to_export == "coaster"){
    if(coaster_pattern_type == "inverted"){
        coaster(true, coaster_d_value);
    } else {
        coaster(false, coaster_d_value);
        if($preview && multicolor_preview == true) color(multicolor_preview_color) coaster(coaster_pattern_type != "inverted" ? true : false, coaster_d_value);
    }

    skirt();
} else if (part_to_export == "both"){
    translate([part_to_export == "both" ? -(shape_radius*2 + border_size + container_wall_buffer + 10)/2 : 0,0,0])
        color(color_of_container) container ();
    translate([part_to_export == "both" ? (shape_radius*2 + border_size + container_wall_buffer + 10)/2 : 0,0,0]) {
        coaster(coaster_pattern_type == "inverted" ? true : false);
        if($preview && multicolor_preview == true) color(multicolor_preview_color) coaster(coaster_pattern_type != "inverted" ? true : false, coaster_d_value);
        skirt();
    }
}

//// Preview Cup/Coasters
// Container/Both Preview
if (part_to_export != "coaster"){
    // Container/Both - Preview Coaster
    if (enable_preview_items == "yes") translate([ part_to_export == "both" ? -(shape_radius*2 + border_size + container_wall_buffer + 10)/2 : 0  ,0,container_base_height]) %coaster_prevew();
}
// Coaster/Both Preview
if (part_to_export != "container") {
    // Coaster/Both - Preview Cup
    if (enable_preview_items == "yes") translate([part_to_export == "both" ? (shape_radius*2 + border_size + container_wall_buffer + 10)/2 : 0,0,coaster_base_height+coaster_pattern_height+0.001]) #%preview_cup();
}

    

////////////////////////////////////////////////////
//////////////////////////////////////////////////// FUNCTIONS
////////////////////////////////////////////////////

function isOdd(x) = (x % 2) == 1;
function radius_from_inradius(i,s) = i / cos(180/s); // i=Inradius | s=Side Count
function inradius_from_radius(r,s) = r * cos(180/s); // r=Radius   | s=Side Count
function radius_from_side_length(l,s) = l / (2*sin(180/s)); // l=Side Length| s=Side Count

////////////////////////////////////////////////////
//////////////////////////////////////////////////// MODULES
////////////////////////////////////////////////////

// Container
module container(){
    // Container Walls
    rotate(rotate_val){
        difference(){
            translate([0,0,container_base_height]) linear_extrude(coaster_height*number_of_coasters) offset(delta=container_wall_width) poly(shape_radius + container_wall_buffer, number_of_sides_val, corner_round);
            frame_space_single();
        }
    }
    
    // Container Base
    linear_extrude(container_base_height){
        rotate(rotate_val) difference(){
            offset(delta=container_wall_width) poly(shape_radius + container_wall_buffer, number_of_sides_val, corner_round);
            offset(delta=-border_size) poly(shape_radius + container_wall_buffer, number_of_sides_val, corner_round);
        }
        intersection(){
            rotate(rotate_val) offset(delta=-border_size) poly(shape_radius + container_wall_buffer + 0.001, number_of_sides_val, corner_round);
            pattern_combine();
        }
    }   
}

///// ### Oi, FIX THIS MESS!!! or don't... it works and you have better things to be doing...
module frame_space_single(){
    translate([0,0,container_base_height-0.001]) linear_extrude(height+0.002) poly(shape_radius + container_wall_buffer, number_of_sides_val, corner_round);
    
    side_length = 2*shape_radius * sin(180/number_of_sides);
    i_radius = inradius_from_radius(shape_radius,number_of_sides);
    height = (coaster_height*number_of_coasters) + 1;
    side_count_val = number_of_sides == 21 ? 8 : number_of_sides;
    rotate_val_01 = isOdd(number_of_sides) ? 180 : 180 + ((360/number_of_sides)/2);
    
    if (container_slot_type == "single") {
        rotate(rotate_val_01) translate([i_radius-1,0,0.001]) rotate([0,90,0]) linear_extrude(container_wall_width*5) hull(){
            translate([-container_base_height-container_slot_corner_val        , -(side_length-container_slot_corner_buffer)/2 +container_slot_corner_val , 0]) circle(container_slot_corner_val); // Bottom Left
            translate([-container_base_height-container_slot_corner_val        ,  (side_length-container_slot_corner_buffer)/2 -container_slot_corner_val , 0]) circle(container_slot_corner_val); // Bottom Right
            translate([-container_base_height-container_slot_corner_val-height , -(side_length-container_slot_corner_buffer)/2 +container_slot_corner_val , 0]) circle(container_slot_corner_val); // Top Left
            translate([-container_base_height-container_slot_corner_val-height ,  (side_length-container_slot_corner_buffer)/2 -container_slot_corner_val , 0]) circle(container_slot_corner_val); // Top Right
        }
    } else {
        step =  container_slot_type == "all" ?          360/side_count_val         : 
                container_slot_type == "alternating" ?  (360/side_count_val) *2    :
                container_slot_type == "double" ?       isOdd(side_count_val) ? 180 + (360/side_count_val)/2: 180                         : 0;
        
        max_value = container_slot_type == "all" ?          360 -(360/side_count_val)  : 
                    container_slot_type == "alternating" ?  isOdd(side_count_val) ? 360 -(360/side_count_val)*2 : 360 -(360/side_count_val)  :
                    container_slot_type == "double" ?       isOdd(side_count_val) ? 180 + (360/side_count_val)/2: 180                         : 0;
        
        for (r = [ 0 : step : max_value ])
        rotate(rotate_val_01+r) translate([i_radius-1,0,0.001]) rotate([0,90,0]) linear_extrude(container_wall_width*5) hull(){
            translate([-container_base_height-container_slot_corner_val        , -(side_length-container_slot_corner_buffer)/2 +container_slot_corner_val , 0]) circle(container_slot_corner_val); // Bottom Left
            translate([-container_base_height-container_slot_corner_val        ,  (side_length-container_slot_corner_buffer)/2 -container_slot_corner_val , 0]) circle(container_slot_corner_val); // Bottom Right
            translate([-container_base_height-container_slot_corner_val-height , -(side_length-container_slot_corner_buffer)/2 +container_slot_corner_val , 0]) circle(container_slot_corner_val); // Top Left
            translate([-container_base_height-container_slot_corner_val-height ,  (side_length-container_slot_corner_buffer)/2 -container_slot_corner_val , 0]) circle(container_slot_corner_val); // Top Right
        }
    }
    
}

// Coaster
module coaster(inverted = false, d=3){
    ///////////////////////// 3D EXPORT
    if(d == 3){
        color(color_of_pattern) translate([0,0,coaster_base_height]){
            // BORDER
            if (inverted == false){
                linear_extrude(coaster_pattern_height+extra_border_height) {
                    rotate(rotate_val) difference(){
                        poly(shape_radius, number_of_sides_val, corner_round);
                        offset(delta=-border_size) poly(shape_radius, number_of_sides_val, corner_round);
                    }
                }
            }
            
            // PATTERN
            linear_extrude(coaster_pattern_height) {
                if (inverted == false){

                    intersection(){
                        rotate(rotate_val) offset(delta=-border_size) poly(shape_radius + 0.001, number_of_sides_val, corner_round);
                        pattern_combine();
                    }
                } else {
                    // INLAY/INVERTED PATTERN
                    intersection(){
                        rotate(rotate_val) offset(delta=-border_size) poly(shape_radius + 0.001, number_of_sides_val, corner_round);
                        difference(){
                            rotate(rotate_val) offset(delta=-border_size) poly(shape_radius + 0.001, number_of_sides_val, corner_round);
                            pattern_combine();
                        }
                    }
                }
            }
        }

        // BASE
        color(color_of_base) if (coaster_base_height != 0 && inverted == false){
           linear_extrude(coaster_base_height) rotate(rotate_val) poly(shape_radius, number_of_sides_val, corner_round);
        }
    } else {
        
        ///////////////////////// 2D EXPORT
        
        color(color_of_pattern) translate([0,0,coaster_base_height]){
            // BORDER
            if (inverted == false){
                    rotate(rotate_val) difference(){
                        poly(shape_radius, number_of_sides_val, corner_round);
                        offset(delta=-border_size) poly(shape_radius, number_of_sides_val, corner_round);
                    }
            }
            
            // PATTERN
            if (inverted == false){
                intersection(){
                    rotate(rotate_val) offset(delta=-border_size) poly(shape_radius + 0.001, number_of_sides_val, corner_round);
                    pattern_combine();
                }
            } else {
                // INLAY/INVERTED PATTERN
                intersection(){
                    rotate(rotate_val) offset(delta=-border_size) poly(shape_radius + 0.001, number_of_sides_val, corner_round);
                    difference(){
                        offset(delta=-border_size) poly(shape_radius + 0.001, number_of_sides_val, corner_round);
                        pattern_combine();
                    }
                }
            }
        }
    }
}

// Preview Coasters
module coaster_prevew(){
    for(n = [0:number_of_coasters-1])
        translate([0,0,n*coaster_height + 0.1]) coaster(false);
}

// Preview Cup
module preview_cup(){
    difference(){
        linear_extrude(5) circle(preview_cup_size/2, $fn = 100);
        translate([0,0,1]) linear_extrude(80) circle((preview_cup_size/2)-5, $fn = 100);
    }
}

/// Draw Text
module pattern_text_lines(){
    line_space_val = enable_custom_line_space == "yes" ? text_line_space + text_thickness : text_size/2 + text_thickness;
    line_count = (len(text_set));
    for (line = [ 0 : 1 : line_count-1 ]){
        line_height_val = text_size + line_space_val;
        temp_value = (line_count-1)/2;
        translate([0,(-text_size/2) + (line_height_val*temp_value) ,0])
        translate([0, (line_height_val * -line) + text_thickness ,0]) 
        text(text = text_set[line], valign = "baseline", halign = "center", size = text_size, font = enable_multiple_fonts == "yes" ? font_set[line] : pattern_text_font_val);
    }
}

/// Transform Text
module pattern_text(text_thickness = 0){
    translate([text_x_nudge,-text_y_nudge,0]) rotate(-text_rotate){
        if(text_thickness == 0){
            pattern_text_lines();
        } else {
            difference(){
                offset(text_thickness){
                    pattern_text_lines();
                }
                pattern_text_lines();
            }
        }
    }
}

/// Draw SVG
module draw_svg(svg_thickness = 0){
    translate([svg_x_nudge,-svg_y_nudge,0]) rotate(-svg_rotate){
        if(svg_thickness == 0){
            scale([file_scale,file_scale]){
                import(file_to_import, center = true);
            }
        } else  {
            difference(){
                offset(svg_thickness){
                    scale([file_scale,file_scale]){
                        import(file_to_import, center = true);
                    }
                }
                scale([file_scale,file_scale]){
                        import(file_to_import, center = true);
                }
            }
        }
    }
}

// Polygon Module
module poly(radius, sides, corner){
    if(corner == 0 || number_of_sides == 21){
        circle(r=radius, $fn = sides);
    } else {
        RoundedRegularPolygon(radius, sides, corner);
    }
}

// Rounded Polygon Module - https://github.com/AVHon/roundedRegularPolygon/blob/master/roundedRegularPolygon.scad
module RoundedRegularPolygon(r, n, c=0){
    cfn_m = $preview ? 1 : 5; // Corner FN multiplyer
	fn = max(n, ($fn==0) ? 360/$fa : $fn); // valid>$fn>$fa
	m = cos(180/n)*r; // Minor radius (origin to edge midpoint)
	i = (m-c)/cos(180/n); // Interior radius (origin to circle center)
	as = (fn-n)/n; // Average number of exposed Sides per corner
	bs = floor(as); // Base number of exposed Sides per corner
	function ms(s) = floor(s*(as-bs)); // Missing Sides by Some corner
	function es(s) = floor(ms(s)-ms(s-1)); // Extra Side on corner?
	hull() for(j=[1:n]) rotate(j*360/n) intersection(){
		mx = m*cos(180/n); my = m*sin(180/n); // edge Midpoint X and Y
		polygon([[0,0],[mx,my],[r,0],[mx,-my]]); // sector of polygon
		s = bs + es(j); // visible Sides this corner will have
		// Not enough edges to round corner; put a point in it.
		if (s<=0) polygon([[0,-0.1*i],[r,0],[0,0.1*i]]);
		// Make a circle to round the corner
		else translate([i,0]) rotate(180/n) circle(c, $fn=s*n*5);
	}
}

// Combine Pattern with Text & SVG
module pattern_combine(){
    if ((used_line_count == 0) && (svg_import_enabled == "no")){
        pattern_select();
    } else {
        difference(){
            pattern_select();
            if ((used_line_count != 0) && (cutout_text == "yes")) {
                pattern_text();
            }
            if ((svg_import_enabled == "yes") && (cutout_svg == "yes")) {
                draw_svg();
            }
        }
        if (used_line_count != 0) { pattern_text(text_thickness); }
        if (svg_import_enabled == "yes") { draw_svg(svg_thickness); }
    }
}

// Skirt for inverted export
module skirt(){
    if(multicolor_export_skirt == "yes"){
        linear_extrude(height = skirt_height) translate([-shape_radius-skirt_buffer_size,0]) ring(shape_radius+skirt_buffer_size, skirt_width);
    }
}

////////////////////////////////////////////////////
//////////////////////////////////////////////////// PATTERN MODULES
////////////////////////////////////////////////////

/// Select Pattern
module pattern_select(){
    rotate(-pattern_rotation) {
        
        size_val = shape_radius * 2;
        
        if(pattern_selection == "honeycomb") {
            honeycomb_pattern(size_val, line_size, line_space);
        }

        if(pattern_selection == "grid") {
            grid_pattern(size_val, line_size, line_space);
        }

        if(pattern_selection == "line") {
            line_pattern(line_size,line_space,size_val);
        }

        if(pattern_selection == "triangle") {
            triangle_pattern(size_val, line_size, line_space);
        }

        if(pattern_selection == "crosshair") {
            crosshair_pattern(size_val, line_size, line_space);
        }
        
        if(pattern_selection == "crossshape") {
            crossshape_pattern(size_val, line_size, line_space);
        }
        
        if(pattern_selection == "crosshex") {
            crosshex_pattern(size_val, line_size, line_space);
        }
        
        if(pattern_selection == "crosstri") {
            crosstri_pattern(size_val, line_size, line_space);
        }

        if(pattern_selection == "square") {
            square_pattern(line_size, line_space, size_val*2);
        }

        if(pattern_selection == "dot") {
            dot_pattern(size_val, line_size, line_space);
        }

        if(pattern_selection == "aperture") {
            aperture_pattern(size_val, line_size, line_space);
        }

        if(pattern_selection == "random") {
            random_pattern(size_val, line_size, line_space);
        }
        
        if(pattern_selection == "zigzag") {
            shape_pattern(size_val, line_size, line_space,4);
        }      

        if(pattern_selection == "trihole") {
            shape_pattern(size_val, line_size, line_space,3);
        }

        if(pattern_selection == "web") {
            spider_web(line_space,size_val,line_size,number_of_support_lines);
        }

        if(pattern_selection == "fancy1") {
            circle_lattice(size_val, line_size, line_space, 4);
        }
        
        if(pattern_selection == "fancy2") {
            circle_lattice(size_val, line_size, line_space, 8);
        }
        
        if(pattern_selection == "fancy") {
            circle_lattice(size_val, line_size, line_space, 40);
        }

        if(pattern_selection == "teardrop"){
            tear_drop_pattern(size_val, line_size, line_space/2);
        }

        if(pattern_selection == "spiral"){
            spiral(size_val, line_size, 100);
        }
        
        if(pattern_selection == "blank"){
            circle(size_val);
        }
        
        if(pattern_selection == "bahtinov"){
            bahtinov_pattern(line_size, line_space, size_val);
        }
        
        if(pattern_selection == "carey"){
            carey_pattern(line_size, line_space, size_val*3);
        }
        
        if(pattern_selection == "cross"){
            cross_pattern(line_size, line_space, size_val*3);
        }
        
        if(pattern_selection == "bahtinov01"){
            bahtinov01_pattern(line_size, line_space, size_val);
        }
        
        if(pattern_selection == "windmill"){
            windmill_pattern(line_size, line_space, size_val);
        }
        
        if(pattern_selection == "none"){
            translate([size_val,size_val,0]) square(1);
        }
        
        if(pattern_selection == "brick"){
            brick_pattern("square",size_val*2);
        }
        
        if(pattern_selection == "brick-rounded"){
            brick_pattern("rounded",size_val*2);
        }
        
        
        if(pattern_selection == "half-loop"){
            half_loop_pattern(line_size, line_space, size_val);
        }
        
        if(pattern_selection == "cross-ring"){
            cross_rings(line_size, line_space, size_val);
        }
        
        if(pattern_selection == "panel-squares"){
            panel_squares_set(size_val, line_size, line_space);
        }
        
        if(pattern_selection == "panel-squares-alt"){
            panel_squares_set01(size_val, line_size, line_space);
        }
        
        if(pattern_selection == "masu-tsunagi"){
            masu_tsunagi_pattern(size_val, line_size, line_space);
        }
        
        if(pattern_selection == "sin-waves"){
            sin_lines_pattern(line_size, line_space, size_val * 1.5);
        }
        
        if(pattern_selection == "crossHearts"){
            heart_pattern(size_val,line_size, line_space);
        }
        
        if(pattern_selection == "shippou-tsunagi"){
            Shippou_Tsunagi(size_val, line_size, line_space);
        }
        
        if(pattern_selection == "rectangles"){
            rectangle_pattern(size_val, line_size, line_space);
        }
        if(pattern_selection == "tsutsuidutsu"){
            tsutsuidutsu_pattern(size_val, line_size, line_space);
        }
        
    }
}

////////////////////////// PATTERNS ////////////////////////////////

module support_lines(size){
    for(rot=[0:90 / number_of_support_lines * 2:180]) {
        rotate(rot) square([size * 2, line_size], center = true);
    } 
}

module support_lines01(size){
    for(rot=[0:360/number_of_sides:360-(360/number_of_sides)]) {
        rotate(rot) translate([0,-line_size/2,0]) square([size * 2, line_size]);
    } 
}

// Grid Pattern
module grid_pattern(size, line_size, line_space) {
    num = ceil(size / (line_size + line_space) * 1.42);
    for(x = [floor(-num / 2) : ceil(num / 2)]) {
        translate([x * (line_size + line_space), 0]) {
            square([line_size, num *(line_size + line_space)], center=true);
        }
        rotate(90) {
            translate([x * (line_size + line_space), 0]) {
                square([line_size, num *(line_size + line_space)], center = true);
            }
        }
    }
}

// Triangle Pattern
module triangle_pattern(size, line_size, line_space) {
    num = ceil(size / (line_size + line_space) * 1.42);
    for(x = [floor(-num / 2):ceil(num / 2)]) {
        translate([x * (line_size + line_space), 0]) {
            square([line_size, num *(line_size + line_space)], center = true);
        }
        rotate(60) {
            translate([x * (line_size + line_space), 0]) {
                square([line_size, num *(line_size + line_space)], center = true);
            }
        }
        rotate(120) {
            translate([x * (line_size + line_space), 0]) {
                square([line_size, num *(line_size + line_space)], center = true);
            }
        }
    }
    
    new_line_space = ( line_space * sin(90)/sin(60) ) + line_size;
    if(centre_item == "yes"){
        rotate(30) circle(new_line_space, $fn = 6);
    }

}

module crosshair_pattern(size, line_size, line_space) {
    line = (line_size + line_space) * 2;
    num = ceil(size / line * 1.42);
    if(centre_item == "yes"){
        circle(d = line + line_size * 2, $fn = ceil(line + line_size * 2));
        support_lines(size*2);
    } else if(centre_item == "no"){
        difference(){
            support_lines(size*2);
            circle(d = line + line_size * 2 -0.001, $fn = ceil(line + line_size * 2));
        }
    } else {
        support_lines(size*2);
    }
    
    for(n = [1:num]) {
        difference() {
            circle(d = n * line + line_size * 2, $fn = ceil(n * line + line_size * 2));
            circle(d = n * line, $fn = ceil(n * line + line_size * 2));
        }
    }
}

module crossshape_pattern(size, line_size, line_space) {
    new_line_size = radius_from_inradius(line_size,number_of_sides);
    new_line_space = radius_from_inradius(line_space,number_of_sides);
    
    line = (new_line_size + new_line_space) * 2;
    num = ceil(size / line * 1.42);

    if(centre_item == "yes"){
        rotate(rotate_val) circle(d = line + new_line_size * 2, $fn = number_of_sides);
        rotate(rotate_val) support_lines01(size);
    } else if(centre_item == "no"){
        difference(){
            rotate(rotate_val) support_lines01(size);
            rotate(rotate_val) circle(d = line + new_line_size * 2 -0.001, $fn = number_of_sides);
        }
    } else {
        rotate(rotate_val) support_lines01(size);
    }
    
    rotate(rotate_val){
        for(n = [1:num]) {
            difference() {
                circle(d = n * line + new_line_size * 2, $fn = number_of_sides);
                circle(d = n * line, $fn = number_of_sides);
            }
        }
    }
}


module crosshex_pattern(size, line_size, line_space) {
    new_line_size = line_size * sin(90)/sin(60);
    new_line_space = line_space * sin(90)/sin(60);
    
    line = (new_line_size + new_line_space) * 2;
    num = ceil(size / line * 1.42);
    
    if(centre_item == "yes"){
        circle(d = line + new_line_size, $fn = 6);
    }
    for(n = [1:num]) {
        difference() {
            circle(d = n * line + new_line_size * 2, $fn = 6);
            circle(d = n * line, $fn = 6);
        }
    }
    // Support Bars
    x = centre_item == "pattern" ? -size : size + new_line_space + new_line_size;
    for(rot=[0:60:300]) {
        rotate(rot) translate([-x,0,0]) square([size * 2, line_size], center = true);
    }
}

module crosstri_pattern(size, line_size, line_space) {
    new_line_size = line_size * sin(90)/sin(30);
    new_line_space = line_space * sin(90)/sin(30);
    
    line = (new_line_size + new_line_space) * 2;
    num = ceil(size / line * 3);
    
    if(centre_item == "yes"){
        circle(d = line + new_line_size * 2, $fn = 3);
    }
    for(n = [1:num]) {
        difference() {
            circle(d = n * line + new_line_size * 2, $fn = 3);
            circle(d = n * line, $fn = 3);
        }
    }
    // Support Bars
    x = centre_item == "pattern" ? -size : -size -new_line_space -new_line_size;
    for(rot=[0:60:300]) {
        rotate(rot) translate([x,0,0]) square([size * 2, line_size], center = true);
    }
}

module honeycomb_pattern(size, line_size, line_space) {
    min_rad = (line_space / 2 * sqrt(3)) / 2 + line_size / 2;
    y_offset = sqrt(min_rad * min_rad * 4 - min_rad * min_rad);
    num_x = ceil(size / min_rad / 2) * 1.42;
    num_y = ceil(size / y_offset) * 1.42;
    difference() {
        square([size * 1.42, size * 1.42], center = true);
        for(y = [floor(-num_y / 2) : ceil(num_y / 2)]) {
            odd = (y % 2 == 0) ? 0 : min_rad;
            for(x = [floor(-num_x / 2) : ceil(num_x / 2)]) {
                translate([x * min_rad * 2 + odd, y * y_offset]) {
                    rotate(30) circle(d=line_space, $fn=6);
                }
            }
        }
    }
}

module dot_pattern(size, line_size, line_space) {
    rad = line_space / 2;
    y_offset = sqrt((rad + line_size / 2) * (rad + line_size / 2) * 4 - (rad + line_size / 2) * (rad + line_size / 2));
    num_x = ceil(size / rad / 2) * 1.42;
    num_y = ceil(size / y_offset) * 1.42;
    difference() {
        square([size * 1.42, size * 1.42], center = true);
        for(y = [floor(-num_y / 2) : ceil(num_y / 2)]) {
            odd = (y % 2 == 0) ? 0 : rad + line_size / 2;
            for(x = [floor(-num_x / 2) : ceil(num_x / 2)]) {
                translate([x * (rad + line_size / 2) * 2 + odd, y * y_offset]) {
                    rotate(30) {
                        circle(d=line_space,$fn = 30);
                    }
                }
            }
        }
    }
}

module aperture_pattern(size, line_size, line_space) {
    circle(d = line_space, $fn = 6); 
    for(rot = [1:2:15]) {
        rotate(360 / 12 * rot) {
            translate([line_space / 2 * cos(360 / 12) - line_size, -line_size]) {
                square([line_size, size]);
            }
        }
    }
}

//////// Shape Patterns
module random_pattern(size, line_size, line_space) {
    diff = (line_space / 10);
    seed=17;
    min_rad = (line_space / 2 * sqrt(3)) / 2 + line_size / 2;
    y_offset = sqrt(min_rad * min_rad * 4 - min_rad * min_rad);
    num_x = ceil(size / min_rad / 2) * 1.42;
    num_y = ceil(size / y_offset) * 1.42;
    difference() {
        square([size * 1.42, size * 1.42], center = true);
        for(y = [floor(-num_y / 2) : ceil(num_y / 2)]) {
            odd = (y % 2 == 0) ? 0 : min_rad;
            for(x = [floor(-num_x / 2) : ceil(num_x / 2)]) {
                translate([x * min_rad * 2 + odd, y * y_offset]) {
                    rotate(30) {
                        single_rand = rands(4,9,1)[0];
                        circle(d=line_space-diff, $fn=single_rand);
                    }
                }
            }
        }
    }
}

module shape_pattern(size, line_size, line_space, fn_pattern) {
    diff = (line_space / 10);
    min_rad = (line_space / 2 * sqrt(3)) / 2 + line_size / 2;
    y_offset = sqrt(min_rad * min_rad * 4 - min_rad * min_rad);
    num_x = ceil(size / min_rad / 2) * 1.42;
    num_y = ceil(size / y_offset) * 1.42;
    difference() {
        square([size * 1.42, size * 1.42], center = true);
        for(y = [floor(-num_y / 2) : ceil(num_y / 2)]) {
            odd = (y % 2 == 0) ? 0 : min_rad;
            for(x = [floor(-num_x / 2) : ceil(num_x / 2)]) {
                translate([x * min_rad * 2 + odd, y * y_offset]) {
                    rotate(30) {
                         circle(d=line_space-diff, $fn=fn_pattern);
                   }
                }
            }
        }
    }
}

//////// Spider web pattern 
module star(radius, wedges){
    //
	angle = 360 / wedges;
	difference() {
		circle(radius, $fn = wedges);
		for(i = [0:wedges - 1]) {
			rotate(angle / 2 + angle * i) translate([radius, 0, 0]) 
			    scale([0.8, 1, 1]) 
				    circle(radius * sin(angle / 2), $fn = 24);
		}
	}
}

module spider_web(ipWebSpacing, strands, ipThickness, wedges) {
    wedges = wedges < 6 ? 6 : wedges;
	for(i = [0:strands - 1]) 
    {
        difference() {
            star(ipWebSpacing * i, wedges);
            offset(r = -ipThickness) star(ipWebSpacing * i, wedges);
        }
	}

	angle = 360 / wedges;
	for(i = [0:wedges - 1])
    {
		rotate(angle * i) translate([0, -ipThickness / 2, 0]) 
			square([ipWebSpacing * strands, ipThickness]);
	}    
}

//////// Fancy Patterns
module circle_lattice(size, line_size, line_space, FN)  {
    for(i = [0,90,180,270]){
        rotate([0,0,i]){
            intersection() {
              square([size,size]); 
              union() {
                for (x=[-line_space:line_space:100+line_space]) {
                  for (y=[-line_space:line_space:100+line_space]){
                     difference()  {
                           translate([x+line_space/2, y+line_space/2]) circle($fn = FN, r=line_space*0.75);
                           translate([x+line_space/2, y+line_space/2]) circle($fn = FN, r=(line_space*0.75)-line_size);
                        }
                    }   // end for y        
                }  // end for x
               } // Union
            }
        }
    }
}

//////// Tear Drop pattern
module tear_drop(r=10){
    union(){
        circle(r=r,$fn=120);
        square([r,r]);
    }
}

module tear_drop_ring(od=10,t=1,so=false){
  difference(){
    tear_drop(r=od/2);
    if(so){
    }else{
        tear_drop(r=od/2-t);
    }
  }
}

module tear_drop_pattern(size, line_size, line_space) {
    line = (line_size + line_space) * 2;
    num = ceil(size / line * 1.42);
    
    if(centre_item == "yes") { 
        rotate(45) tear_drop_ring(line, line_size, true);
    }
    
    if(centre_item == "no") {
        difference(){
            support_lines(size*2);
            tear_drop_ring(line, line_size, true);
        }
    } else {
        support_lines(size*2);
    }
    
    rotate(45) for(n = [1:num]) {
        tear_drop_ring(n * line, line_size);
    }
}

//////// Spiral Pattern
module ring (size, line_size, fn){
    $fn = 90;
    translate([size,0]){
        difference(){
            circle(size,$fn = fn);
            circle(size-line_size,$fn = fn);
        }
    }
};

module ring_slice (size, line_size, fn){
    difference(){
        ring(size, line_size, fn);
        translate([size,0]){
            for(i = [180:90:360]) {
                rotate([0,0,i]){
                    square(size);
                }
            }
        }
    }
}

module spiral(size, line_size, fn){
    size = size * 1.1;
    if(centre_item == "yes"){
        circle(size/10,$fn = fn);
    } else {
        difference(){
            circle(size/10,$fn = fn);
            circle(size/10-line_size,$fn = fn);
        }
    }
    deg = (360/number_of_support_lines);
    for(n = [0:deg:360]) {
        rotate([0,0,n+7]){
            translate([-size/10,0]) ring_slice (size/2, line_size, fn);
        }
    }
}

// Rings
module rings(size, line_size, fn){
    size = size * 1.1;
    if(centre_item == "yes"){
        circle(size/10,$fn = fn);
    } else {
        difference(){
            circle(size/10,$fn = fn);
            circle(size/10-line_size,$fn = fn);
        }
    }
    deg = (360/number_of_support_lines);
    for(n = [0:deg:360]) {
        rotate([0,0,n]){
            translate([-size/10,0]) ring(size/2, line_size, fn);
        }
    }
}

//////// Bars module
module bars(line_size, line_space, size, num=5) {
    num = round(num);
    for (i=[-num:num]) {
        translate([size/4,i*2*line_space]) square([size/2,line_size], center=true);
    }
    
}

//////// Line Pattern (New Version)
module line_pattern(line_size, line_space, size) {
    
    translate([-size,0,0]) {
    
        // Vertical Support Lines
        numBarsVert = size*2/(number_of_support_lines+1);
        for(i = [numBarsVert:numBarsVert:size*2-0.001]) {
            translate([i,0]) square([line_size,size*2], center=true);
        }
    }
        
    // Horizontal Support Lines
    numBars = size / line_space / 2 -1;
    translate([-size,0,0])
        bars(line_size, line_space, size*4, numBars);

}



//////// Bahtinov Pattern
module bahtinov_pattern(line_size, line_space, size) {
    size_double = size * 2;
    rotate([0,0,90]) square([size_double, line_size], center=true);
    rotate([0,0,0]) square([size_double, line_size], center=true);
    numBars = size / line_space / 2 -1;
    // +X +Y bars
    intersection() {
        rotate([0,0,30])
            bars(line_size, line_space, size_double, numBars);
        square([size_double, size_double], center=false);
    }
    // +X -Y bars
    intersection() {
        rotate([0,0,-30])
            bars(line_size, line_space, size_double, numBars);
        translate([0,-size_double])
        square([size_double, size_double], center=false);
    }
    // -X bars
    rotate([0,0,180])
        bars(line_size, line_space, size_double, numBars);
    
    
    
}

//////// Bahtinov Alt Pattern
module bahtinov01_pattern(line_size, line_space, size) {
    number_of_support_lines = number_of_support_lines < 1 ? 1 : number_of_support_lines;
    size_double = size * 2;

    for(rot=[90:90:180]) {
        rotate(rot) square([size_double * 3, line_size], center = true);
    } 

    numBars = size / line_space / 2 -1;
    for(i = [0:180:180]) {
        rotate(i){
            // +X +Y bars
            intersection() {
                rotate([0,0,45]) bars(line_size, line_space, size_double, numBars);
                square([size_double, size_double], center=false);
            }
            // +X -Y bars
            intersection() {
                bars(line_size, line_space, size_double, numBars);
                translate([0,-size_double]) square([size_double, size_double], center=false);
            }
        }
    }
}

////////// Carey Pattern
module carey_pattern (line_size, line_space, size) {
    rotate([0,0,90]) square([size, line_size], center=true);
    numBars = size/2 / line_space / 2 -1;
    
    for(i = [0:180:180]) {
        rotate(i){
    
            // +X +Y bars
            intersection() {
                rotate([0,0,10]) bars(line_size, line_space, size, numBars);
                difference(){
                    square([size, size], center=false);
                    rotate([0,0,10]) translate([0,-size-line_size]) square([size, size], center=false);
                }
            }
            // +X -Y bars
            intersection() {
                rotate([0,0,-10]) bars(line_size, line_space, size, numBars);
                difference(){
                    translate([0,-size]) square([size, size], center=false);
                    rotate([0,0,-10]) translate([0,line_size]) square([size, size], center=false);
                }
            }
        }
    }
}

//////// Cross Pattern
module cross_pattern(line_size, line_space, size) {
    number_of_support_lines = number_of_support_lines > 2 ? 2 : number_of_support_lines;
    if(number_of_support_lines != 0){
        for(rot=[0:90 / number_of_support_lines * 2:180]) {
            rotate(rot) square([size * 3, line_size], center = true);
        } 
    }
    numBars = size/2 / line_space / 2 -1;
    for(i = [0:90:270]) {
        rotate(i){
            intersection() {
                rotate([0,0,45])
                    bars(line_size, line_space, size, numBars);
                square([size, size], center=false);
            }
        }
    }
}

//////// Square (New Version)
module square_pattern(line_size, line_space, size) {
    number_of_support_lines = number_of_support_lines < 1 ? 1 : number_of_support_lines;

    for(rot=[90:90:180]) {
        rotate(rot) square([size * 3, line_size], center = true);
    } 

    numBars = size/2 / line_space / 2 -1;
    for(i = [0:90:270]) {
        rotate(i){
            // +X +Y bars
            intersection() {
                rotate([0,0,-45]) translate([-size/4,0]) bars(line_size, line_space, size, numBars);
                square([size, size], center=false);
            }
        }
    }
}

//////// Windmill Pattern
module windmill_pattern(line_size, line_space, size) {
    
    number_of_support_lines = (number_of_support_lines/2==round(number_of_support_lines/2)) ? number_of_support_lines : number_of_support_lines - 1;
    
    number_of_support_lines_val = number_of_support_lines < 4 ? 4 : number_of_support_lines;
    
    angle_val = 360 / number_of_support_lines_val;

    size_double = size * 2;
    
    // Support Lines
    for(rot=[angle_val:angle_val:360]) {
        rotate(rot) square([size_double * 3, line_size], center = true);
    } 
    
    // Pattern
    numBars = size / line_space / 2 -1;
    for(i = [angle_val:angle_val:360]) {
        rotate(i){
            // +X +Y bars
            intersection() {
                rotate([0,0,0]) bars(line_size, line_space, size_double, numBars);
                //square([size, size], center=false);
                hull(){
                    translate([line_size/2,-line_size]) square([size_double, line_size]);
                    rotate(-angle_val) translate([size/4,line_size/2]) square([size/2, line_size]);
                }
            }

        }
    }
}


//////// Brick Pattern

module brick(brick_width, shape){
    if(shape == "square"){
        square([brick_width,line_space], true);
    } else {
        hull(){
            fn_val = line_space < 8 ? 20 : line_space*2;
            translate([(-brick_width/2)+(line_space/2),0,0]) circle(line_space/2,$fn = fn_val);
            translate([(brick_width/2)-(line_space/2),0,0]) circle(line_space/2,$fn = fn_val);
        }
    }
}

module brick_line(number_of_support_lines, brick_width, shape, size_val){
    val = ((brick_width) + line_size); // X Loop
    val01 = (line_space+line_size)*2; // Y Loop
    numbar = 20; // ### Fix this
    
    for (x=[-numbar*val:val:numbar*val], y=[-numbar*val01:val01:numbar*val01]) 
        translate([x,y,0])
            brick(brick_width, shape);
}

module brick_pattern(shape="square",size_val){
    number_of_support_lines = number_of_support_lines < 2 ? 1 : number_of_support_lines/2;
    brick_width = line_space * number_of_support_lines;
    difference(){
        square([size_val,size_val], center = true);
        union(){
            brick_line(number_of_support_lines, brick_width, shape, size_val);
            translate([((brick_width) + line_size)/2, line_space+line_size, 0]) brick_line(number_of_support_lines, brick_width, shape);
        }
    }
}


////////////// Half Loop
module half_loop(line_size, line_space){
    intersection(){
        translate([-line_space/2,0,0]) square(line_space);
        difference(){
            circle(line_space/2, $fn = line_space*4);
            circle(line_space/2 -line_size, $fn = line_space*4);
        }
    }
}


module half_loop_pattern(line_size, line_space, size){
    line_space = line_space*10;
    space = (line_space-line_size);
    num = ceil(size / space);
    for (i = [ -num * space : space : num * space ] , n = [ -num * space/2 : space/2 : num * space/2 ] ){
        translate([i,n*2,0]) half_loop(line_size, line_space);
        translate([space/2 + i, space/2 + n*2,0]) half_loop(line_size, line_space);
    }
    
    if(centre_item == "yes"){
        circle(line_space/2);
    } else if (centre_item == "pattern") {
        
        for (i = [ -num * space : space : num * space ] , n = [ -num * space/2 : space/2 : num * space/2 ] ){
            translate([i*2,n*4,0]) circle(line_space/2);
        }
    }
    
    
}

////////////// Cross Ring
module cross_ring_loop(line_size, line_space){
    difference(){
        circle(line_space/2, $fn = line_space*4);
        circle(line_space/2 -line_size, $fn = line_space*4);
    }
    for (r=[0:90:90]) rotate(r) translate([0,line_space - line_size/2,0]) square([line_size,line_space],center=true);
}

module cross_ring_circle(line_size, line_space){
    circle(line_space/2, $fn = line_space*4);
    for (r=[0:90:90]) rotate(r) translate([0,line_space - line_size/2,0]) square([line_size,line_space],center=true);
}

module cross_ring_cross(line_size, line_space){
    difference(){
        circle(line_space/2, $fn = line_space*4);
        circle(line_space/2 -line_size, $fn = line_space*4);
    }
    for (r=[0:90:90]) rotate(r) translate([0,line_size/2,0]) square([line_size,line_space*2],center=true);
}


module cross_rings(line_size, line_space, size){
    space = line_space*2 - line_size;
    num = ceil(size/space);
    for (i = [ -num * space : space : num * space ] , n = [ -num * space : space : num * space ] ){
        if(centre_item == "yes"){
            translate([i,n,0]) cross_ring_circle(line_size, line_space);
        } else if (centre_item == "no") {
            translate([i,n,0]) cross_ring_loop(line_size, line_space);
        } else {
            translate([i,n,0]) cross_ring_cross(line_size, line_space);
        }
    }
}

//// Square Panels

module panel_squares(line_size, line_space){
    square_size = line_size*3 + line_space*3;
    difference(){
        square([square_size , square_size], center = true);
        square([square_size - line_size , square_size - line_size], center = true);
    }

    for(i = [-1:2:1]){
        translate([0,i*(line_size/2 + line_space/2),0]) square([square_size-line_size , line_size], center = true);

    }
}

module panel_squares_set(size, line_size, line_space){
    square_size = line_size*3 + line_space*3;
    num = ceil(size / square_size);
    for(i = [-num:2:num], n = [-num:2:num]){
        translate([i*square_size,n*square_size,0]) panel_squares(line_size, line_space);
        translate([(i *square_size) + square_size,(n *square_size) + square_size,0]) panel_squares(line_size, line_space);
    }
    
    translate([square_size,0,0]) for(i = [-num:2:num], n = [-num:2:num]){
        translate([i*square_size,n*square_size,0]) rotate(90) panel_squares(line_size, line_space);
        translate([(i *square_size) + square_size,(n *square_size) + square_size,0]) rotate(90) panel_squares(line_size, line_space);
    }
}

// Alt Squares
module panel_squares01(line_size, line_space){
    square_size = line_size*3 + line_space*3;
    difference(){
        square([ square_size , square_size ], center = true);
        square([ square_size - line_size , square_size - line_size ], center = true);
    }
    for(i = [0:3]){
        rotate([0,0,i*90]) translate([-line_space/2, -line_size -line_space/2 ,0]) square([ line_size + line_space*2 , line_size ]);
    }
}

module panel_squares_set01(size, line_size, line_space){
    square_size = line_size*3 + line_space*3;
    num = ceil(size / square_size);
    for(i = [-num:1:num], n = [-num:1:num]){
        translate([i*square_size,n*square_size,0]) panel_squares01(line_size, line_space);
    }
}

// Panel | Masu-Tsunagi
module masu_tsunagi_square(line_size, line_space){
    square_size = line_size*3 + line_space*3;
    difference(){
        square([square_size , square_size], center = true);
        square([square_size - line_size , square_size - line_size], center = true);
    }
    
    if((centre_item == "yes") || (centre_item == "pattern")){
        rotate(45) square([line_space*1.5 + line_size*2, line_space*1.5 + line_size*2], center = true);
    } else {
        rotate(45) difference(){
            square([line_space*1.5 + line_size*2, line_space*1.5 + line_size*2], center = true);
            square([line_space*1.5, line_space*1.5], center = true);
        }
    }
    
    intersection(){
        square([square_size - line_size , square_size - line_size], center = true);
        for(i = [0:3]){
            rotate(i*90) translate([(line_space*1.25 + line_size) * sqrt(2),0,0]) rotate(45) {
                if(centre_item == "pattern"){
                    square([line_space + line_size*2, line_space + line_size*2], center = true);
                } else {
                    difference(){
                        square([line_space + line_size*2, line_space + line_size*2], center = true);
                        square([line_space, line_space], center = true);
                    }
                }
            }
        }
    }
}

module masu_tsunagi_pattern(size, line_size, line_space){
    square_size = line_size*3 + line_space*3;
    num = ceil(size / square_size);
    for(i = [-num:1:num], n = [-num:1:num]){
        translate([i*square_size,n*square_size,0]) masu_tsunagi_square(line_size, line_space);
    }
}


///// Sin Lines
module sin_line(line_size, size){
    step = $preview ? 0.5 : 0.05;
    for(i=[ -size/2 : step : size/2 ]){
        hull(){
            translate([i, sin(i*6)*2, 0]) circle(line_size/2);
            translate([i+step, sin(i*6)*2, 0]) circle(line_size/2);
        }
    }
}

module sin_lines_pattern(line_size, line_space, size){
    num = ceil(size/2 / (line_space + line_size)) ;
    for (i = [-num:num]){
        translate([ 0, (line_space + line_size) * i, 0 ]) sin_line(line_size, size);
    }
    
}

//// Cross-Heart
module heart(size){
    translate([0,-size/5,0]){
        rotate(45){
            square(size, center = true);
            translate([0,size/2,0]) circle(size/2);
            translate([size/2,0,0]) circle(size/2);
        }
    }
}

module heart_ring(size, line_size){
    difference(){
        heart(size);
        offset(delta=-line_size) heart(size);
    }
}



module heart_pattern(size, line_size, line_space){
    line = (line_size + line_space) * 2;
    num = ceil(size / line * 1.42);
    
    if(centre_item == "yes"){
        heart(line + line_size * 2);
    } else {
        heart_ring(line + line_size * 2, line_size);
    }
    
    for(n = [2:num]) {
        heart_ring(n * line + line_size * 2, line_size);
    }
    
    // Support Lines
    if((centre_item == "yes") || (centre_item == "pattern")){
        support_lines(size*2);
    } else {
        difference(){
            support_lines(size*2);
            heart(line + line_size * 2 -0.001);
        }
    }
}

////////////////
module full_ring(line_size, line_space){
    difference(){
        circle(line_space + line_size, $fn = (line_space*6 < 20 ? 20 : line_space*6));
        circle(line_space, $fn = (line_space*6 < 20 ? 20 : line_space*6) );
    }
}

module Shippou_Tsunagi(size, line_size, line_space){
    ring_size = line_space + line_size;
    num = ceil(size*0.6 / (line_space + line_size)) ;
    for (i = [-num:num], x = [-num:num]){
        translate([             (ring_size*2)*i,             (ring_size*2)*x, 0 ]) full_ring(line_size, line_space);
        translate([ ring_size + (ring_size*2)*i, ring_size + (ring_size*2)*x, 0 ]) full_ring(line_size, line_space);
    }
}

/////////////////
module rectangle(line_size, line_space){
    difference(){
        square([line_space + line_size*2, line_space*3 + line_size*2.001], center = true);
        square([line_space              , line_space*3 + line_size*2], center = true);
    }
}

module rectangle_pattern(size, line_size, line_space){

    num = ceil(size*0.6 / (line_space + line_size)/4) ;
    for (i = [-num:num], x = [-num:num]){
        translate([(line_space*4 + line_size*4)*i ,    (line_space*4 + line_size*4) * x    ,0]) rotate(90)  rectangle(line_size, line_space);
        
        translate([ (line_space*2 + line_size*2) + ((line_space*4 + line_size*4)*i) , (line_space*2 + line_size*2) + ((line_space*4 + line_size*4) * x) ,0]) rotate(90)  rectangle(line_size, line_space);
        
        translate([(line_space*4 + line_size*4) * i , (line_space*2 + line_size*2) + ((line_space*4 + line_size*4) * x) ,0])  rectangle(line_size, line_space);
        
        translate([0,0,0]) translate([(line_space*2 + line_size*2) + ((line_space*4 + line_size*4) * i) , (line_space*4 + line_size*4) + ((line_space*8 + line_size*8) * x/2) ,0])  rectangle(line_size, line_space);
        
    }
}

////////////////
module tsutsuidutsu_square(line_size, line_space){
    square_size = line_size*3 + line_space*2;
    
    difference(){
        square([square_size , square_size], center = true);
        square([square_size - line_size , square_size - line_size], center = true);
    }
    
    if(centre_item == "yes"){
        square([line_space + line_size*2, line_space + line_size*2], center = true);
    } else {
        difference(){
            square([line_space + line_size*2, line_space + line_size*2], center = true);
            square([line_space, line_space], center = true);
        }
    }
    
    if(centre_item == "no"){
        for(i = [1:4]){
            rotate(45 + i*90) translate([(line_space + line_size) * sqrt(2),0,0])
                square([(line_space + line_size*2) * sqrt(2), line_size], center = true);
        }
    } else {
        for(i = [1:2]){
            rotate(45 + i*90)
                square([(line_space*2 + line_size*2) * sqrt(2), line_size], center = true);
        }
    }
}

module tsutsuidutsu_pattern(size, line_size, line_space){
    square_size = line_size*3 + line_space*2;
    num = ceil(size / square_size);
    for(i = [-num:1:num], n = [-num:1:num]){
        translate([i*square_size,n*square_size,0]) tsutsuidutsu_square(line_size, line_space);
    }
}
